/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  My algorithmic reference was the filter code in the GLPLed spkit.
*/

/*

adjust needs to check filter indices for accuracy

prolly want master level also

add BANDREJECT

NB: performance improvements from template-based variants of this
  are lost in the noise..the next step is probably generating filter code
  from specs..code that uses fixed size param structures etc

*/

extern "C"
{
#include <math.h>
}

#ifndef M_PI
#define M_PI 3.14159
#endif

#include "ButterJam.h"

void
ButterJam::adjust(size_t channel, size_t filter, float level)
{
  filters[channel][filter].level = level;
}

void
ButterJam::adjust(size_t channeln, size_t filtern, float level,
	      enum type t, float freq, float q=0)
{
  filter *f = &filters[channeln][filtern];
  f->adjust(sr, t, freq, q);
  f->level = level;
}

void
ButterJam::filter::adjust(float sr, enum type newt, float f, float q=0)
{
  t = newt;

  switch (t)
    {
    case NONE: 
      // make SURE this never outputs anything
      level = 0;
      break;
    case LOWPASS:
      {
	float C = 1.0 / tan(M_PI * f / sr);

	a[0] = 1.0 / (1.0 + sqrt(2.0) * C + C * C);
	a[1] = 2.0 * a[0];
	a[2] = a[0];
	
	b[0] = 2 * (1.0 - C * C) * a[0];
	b[1] = (1.0 - sqrt(2.0) * C + C * C) * a[0];
      }
      break;
    case BANDPASS:
      {
	float C = 1.0 / tan(M_PI * q / sr);
	float D = 2 * cos(2 * M_PI * f / sr);

	a[0] = 1.0 / (1.0 + C);
	a[1] = 0.0;
	a[2] = -a[0];

	b[0] = -C * D * a[0];
	b[1] = (C - 1.0) * a[0];
      }
      break;
    case HIGHPASS:
      {
	float C = tan(M_PI * f / sr);

	a[0] = 1.0 / (1.0 + sqrt(2.0) * C + C * C);
	a[1] = -2.0 * a[0];
	a[2] = a[0];

	b[0] = 2 * (C * C - 1.0) * a[0];
	b[1] = (1.0 - sqrt(2.0) * C + C * C) * a[0];
      }
      break;
    }
}

ButterJam::ButterJam(size_t nchannels, size_t nfilters, float sr)
: nchannels(nchannels), nfilters(nfilters), sr(sr)
{
  // create array of filters
  filters = new filter *[nchannels];
  for (size_t i = 0; i < nchannels; i++)
    filters[i] = new filter[nfilters];
}

void
ButterJam::go(float *in, float *out, size_t nsamps)
{
  for (size_t i = 0; i < nsamps; i++)
    {
      for (size_t j = 0; j < nchannels; j++)
	{
	  float v = 0;
	  float in1 = *in;
	  
	  for (size_t k = 0; k < nfilters; k++)
	    {
	      filter *f = &filters[j][k];
	      
	      float out1 = f->a[0] * in1 + f->a[1] * f->x[0] 
		+ f->a[2] * f->x[1]
		- f->b[0] * f->y[0] - f->b[1] * f->y[1];
	      
	      f->x[1] = f->x[0];
	      f->x[0] = in1;
	      f->y[1] = f->y[0];
	      f->y[0] = out1;
	      
	      v += out1 * f->level;
	    }
	  
	  in++;
	  *out++ = v;
	}
    }
}

//#define TEST
#ifdef TEST

float buf[1024];
float buf1[1024];

int
main()
{
  ButterJam bj(2, 3, 44100);
  bj.adjust(0, 0, 1.0, ButterJam::LOWPASS, 250);
  bj.adjust(1, 0, 1.0, ButterJam::LOWPASS, 250);
  bj.adjust(0, 1, 1.0, ButterJam::BANDPASS, 1250, 2500);
  bj.adjust(1, 1, 1.0, ButterJam::BANDPASS, 1250, 2500);
  bj.adjust(0, 2, 1.0, ButterJam::HIGHPASS, 2500);
  bj.adjust(1, 2, 1.0, ButterJam::HIGHPASS, 2500);

  for (size_t i = 0; i < 861; i++)
    bj.go(buf, buf1, 512);

  return 0;
}
#endif
